import argparse
import json
import logging
import logging.config
import time
from pathlib import Path

from open_dread_rando import dread_patcher


def create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("--input-path", required=True, type=Path,
                        help="Path to where the extracted Metroid Dread romfs to randomize can be found.")
    parser.add_argument("--output-path", required=True, type=Path,
                        help="Path to where the modified files will be written to.")
    parser.add_argument("--input-json", required=True, type=Path,
                        help="Path to the configuration json.")
    parser.add_argument("-q", "--quiet", action="store_true", help="Disables all info and debug logs.")
    return parser


def setup_logging():
    handlers = {
        'default': {
            'level': 'DEBUG',
            'formatter': 'default',
            'class': 'logging.StreamHandler',
            'stream': 'ext://sys.stdout',  # Default is stderr
        },
    }
    logging.config.dictConfig({
        'version': 1,
        'formatters': {
            'default': {
                'format': '[%(asctime)s] [%(levelname)s] [%(name)s] %(funcName)s: %(message)s',
            }
        },
        'handlers': handlers,
        'disable_existing_loggers': False,
        'loggers': {
            'default': {
                'level': 'DEBUG',
            },
        },
        'root': {
            'level': 'DEBUG',
            'handlers': list(handlers.keys()),
        },
    })


def main():
    setup_logging()
    parser = create_parser()
    args = parser.parse_args()
    if args.quiet:
        logging.getLogger().setLevel(logging.WARNING)

    with args.input_json.open() as f:
        configuration = json.load(f)

    start = time.time()
    dread_patcher.patch_extracted(
        args.input_path,
        args.output_path,
        configuration,
    )
    end = time.time()
    print(f"Patcher took {end - start:.03f} seconds")



import dataclasses
import typing

from mercury_engine_data_structures.formats.bsmat import Bsmat

from open_dread_rando.patcher_editor import PatcherEditor


@dataclasses.dataclass(frozen=True)
class MaterialData:
    base_mat: str # path to the "base" material
    new_mat_name: str = None # the name of the new material
    new_path: str = None # the new path for the material, none if its only modifiying an existing mat
    uniform_params: dict[str, typing.Sequence] = None # a dictionary of uniform params stored in shader_stages[0]
    sampler_params: dict[str, dict[str, typing.Union[str, int, float, list]]] = None
    # a dictionary of sampler params stored in shader_stages[0]
    # key is the name of the sampler param
    # value is a dict of key/value pairs for keys of the sampler (i.e. { "filepath": "path/to/texture.bsmat" })


def create_custom_material(editor: PatcherEditor, material_data: MaterialData) -> Bsmat:
    mat = editor.get_parsed_asset(material_data.base_mat, type_hint=Bsmat)

    # update name
    if material_data.new_mat_name:
        mat.raw.name = material_data.new_mat_name

    # update uniforms
    if material_data.uniform_params:
        for uni_name, uni_value in material_data.uniform_params.items():
            mat.get_uniform(uni_name).value = uni_value

    # update samplers
    if material_data.sampler_params:
        for sam_name, sam_dict in material_data.sampler_params.items():
            if "filepath" in sam_dict:
                editor.check_file_exists(sam_dict["filepath"], "Custom material missing sampler param")

            sampler = mat.get_sampler(sam_name)
            for sam_key, sam_value in sam_dict.items():
                sampler[sam_key] = sam_value

    if material_data.new_path:
        editor.add_new_asset(material_data.new_path, mat, [])

    return mat



import dataclasses
import functools
from enum import Enum

from mercury_engine_data_structures.formats import Bmsad
from mercury_engine_data_structures.formats.bmsad import ActorDefFunc

from open_dread_rando.files import templates_path
from open_dread_rando.misc_patches.material_patcher import MaterialData, create_custom_material
from open_dread_rando.misc_patches.model_patcher import ModelData, create_custom_model
from open_dread_rando.patcher_editor import PatcherEditor

MISSILE_MDL = "actors/props/doorshieldmissile/models/doorshieldmissile.bcmdl"
SUPER_MDL = "actors/props/doorshieldsupermissile/models/doorshieldsupermissile.bcmdl"
ENERGY_MDL = "actors/props/door_shield_plasma/models/doorshieldplasma.bcmdl"
MISSILE_MAT = "actors/props/doorshieldmissile/models/imats/doorshieldmissile_mp_opaque_01.bsmat"
SUPER_MAT = "actors/props/doorshieldsupermissile/models/imats/doorshieldsupermissile_mp_opaque_01.bsmat"
ENERGY_MP_MAT = "actors/props/door_shield_plasma/models/imats/doorshieldplasma_mp_opaque_01.bsmat"
ENERGY_FX_MAT = "actors/props/door_shield_plasma/models/imats/doorshieldplasma_matfx.bsmat"
SMOOTH_NORMALS = "actors/props/doorshieldmissile/models/textures/shield_no_normals_nm.bctex"
SMOOTH_ATTRIBUTES = "actors/props/doorshieldmissile/models/textures/shield_no_attribs_at.bctex"


@functools.cache
def _template_read_shield(file: str) -> bytes:
    return templates_path().joinpath(file).read_bytes()


class DoorTemplates(Enum):
    HEXAGONS = "template_doorshield_hexs.bmsad"
    TRIANGLES = "template_doorshield_tris.bmsad"
    ENERGY = "template_doorshield_energy.bmsad"


@dataclasses.dataclass(frozen=True)
class ShieldData:
    name: str
    type: DoorTemplates  # TODO make these modify the in-game bmsad's so we don't need templates
    weaknesses: list[str]
    actordef: str
    default_mdl: ModelData  # the default new model
    default_mats: list[MaterialData]  # a list of materials needed for the default model
    alternate_mats: list[MaterialData]  # a list of materials needed for the alternate model
    alternate_mdl: ModelData | None = None  # if none, ALTERNATE shields use default model and only change materials
    collision: str | None = None


ALL_SHIELD_DATA: dict[str, ShieldData] = {
    "ice_missile": ShieldData(
        name="doorshieldicemissile",
        type=DoorTemplates.HEXAGONS,
        weaknesses=["ICE_MISSILE"],
        actordef="actors/props/doorshieldicemissile/charclasses/doorshieldicemissile.bmsad",
        default_mdl=ModelData(
            base_model=SUPER_MDL,
            new_path="actors/props/doorshieldicemissile/models/doorshieldicemissile.bcmdl",
            materials={
                "mp_opaque_01": "actors/props/doorshieldicemissile/models/imats/doorshieldicemissile_mp_opaque_01.bsmat"
            }
        ),
        default_mats=[
            MaterialData(
                base_mat=MISSILE_MAT,
                new_mat_name="doorshieldicemissile_mp_opaque_01",
                new_path="actors/props/doorshieldicemissile/models/imats/doorshieldicemissile_mp_opaque_01.bsmat",
                uniform_params={
                    "vConstant0": [0.0, 0.94, 0.95, 1.0]
                }
            )
        ],
        alternate_mats=[
            MaterialData(
                base_mat=MISSILE_MAT,
                new_mat_name="doorshieldicemissile_mp_opaque_01",
                new_path="actors/props/doorshieldicemissile/models/imats/doorshieldicemissile_mp_opaque_01.bsmat",
                uniform_params={
                    "vConstant0": [0.0, 0.94, 0.95, 1.0]
                },
                sampler_params={
                    "texBaseColor": {"filepath": ("actors/props/doorshieldicemissile/models"
                                                  "/textures/doorshieldicemissile_alt_bc.bctex")},
                    "texAttributes": {"filepath": SMOOTH_ATTRIBUTES},
                    "texNormals": {"filepath": SMOOTH_NORMALS}
                }
            )
        ]
    ),

    "diffusion_beam": ShieldData(
        name="shield_diffusion",  # NOTE required to be <= len("doorshieldplasma") until bcmdl can alter pointers
        type=DoorTemplates.ENERGY,
        weaknesses=["DIFFUSION_BEAM"],
        actordef="actors/props/shield_diffusion/charclasses/shield_diffusion.bmsad",
        default_mdl=ModelData(
            base_model=ENERGY_MDL,
            new_path="actors/props/shield_diffusion/models/shield_diffusion.bcmdl",
            materials={
                "matfx": "actors/props/shield_diffusion/models/imats/shield_diffusion_matfx.bsmat",
                "mp_opaque_01": "actors/props/shield_diffusion/models/imats/shield_diffusion_mp_opaque_01.bsmat"
            }
        ),
        alternate_mdl=ModelData(
            base_model=SUPER_MDL,
            new_path="actors/props/shield_diffusion/models/shield_diffusion.bcmdl",
            materials={
                "mp_opaque_01": "actors/props/shield_diffusion/models/imats/shield_diffusion_mp_opaque_01.bsmat"
            }
        ),
        default_mats=[
            # matfx
            MaterialData(
                base_mat=ENERGY_FX_MAT,
                new_mat_name="shield_diffusion_matfx",
                new_path="actors/props/shield_diffusion/models/imats/shield_diffusion_matfx.bsmat",
                uniform_params={
                    "vConstant0": [1.0, 0.05, 0.05, 1.0]
                }
            ),
            # mp_opaque_01
            MaterialData(
                base_mat=ENERGY_MP_MAT,
                new_mat_name="shield_diffusion_mp_opaque_01",
                new_path="actors/props/shield_diffusion/models/imats/shield_diffusion_mp_opaque_01.bsmat",
                uniform_params={
                    "vConstant1": [1.0, 0.05, 0.05, 0.0]
                }
            )
        ],
        alternate_mats=[
            # mp_opaque_01
            MaterialData(
                base_mat=SUPER_MAT,
                new_mat_name="shield_diffusion_mp_opaque_01",
                new_path="actors/props/shield_diffusion/models/imats/shield_diffusion_mp_opaque_01.bsmat",
                uniform_params={
                    # "fAlbedoEmissiveFactor": [2.0],
                    "vAlbedoEmissiveColorMultiplier": [1.0, 0.05, 0.05, 1.0]
                },
                sampler_params={
                    "texBaseColor": {"filepath": ("actors/props/shield_diffusion/models/"
                                                  "textures/shield_diffusion_alt_bc.bctex")},
                    "texAttributes": {"filepath": SMOOTH_ATTRIBUTES},
                    "texNormals": {"filepath": SMOOTH_NORMALS}
                }
            )
        ]
    ),

    "storm_missile": ShieldData(
        name="doorshieldstormmissile",
        type=DoorTemplates.TRIANGLES,
        weaknesses=["MULTI_LOCKON_MISSILE"],
        actordef="actors/props/doorshieldstormmissile/charclasses/doorshieldstormmissile.bmsad",
        default_mdl=ModelData(
            base_model=SUPER_MDL,
            new_path="actors/props/doorshieldstormmissile/models/doorshieldstormmissile.bcmdl",
            materials={
                "mp_opaque_01": ("actors/props/doorshieldstormmissile/models"
                                 "/imats/doorshieldstormmissile_mp_opaque_01.bsmat")
            }
        ),
        default_mats=[
            MaterialData(
                base_mat=SUPER_MAT,
                new_mat_name="doorshieldstormmissile_mp_opaque_01",
                new_path="actors/props/doorshieldstormmissile/models/imats/doorshieldstormmissile_mp_opaque_01.bsmat",
                uniform_params={
                    "fAlbedoEmissiveFactor": [2.0],
                    "vAlbedoEmissiveColorMultiplier": [1.0, 1.0, 1.0, 1.0],
                    "vConstant0": [7.0, 0.15, 0.0, 1.0]
                },
                sampler_params={
                    "texBaseColor": {"filepath": ("actors/props/doorshieldstormmissile/models"
                                                  "/textures/doorshieldstormmissile_bc.bctex")},
                    "texAttributes": { "filepath": SMOOTH_ATTRIBUTES },
                    "texNormals": { "filepath": SMOOTH_NORMALS }
                }
            )
        ],
        alternate_mats=[
            MaterialData(
                base_mat=SUPER_MAT,
                new_mat_name="doorshieldstormmissile_mp_opaque_01",
                new_path="actors/props/doorshieldstormmissile/models/imats/doorshieldstormmissile_mp_opaque_01.bsmat",
                uniform_params={
                    "fAlbedoEmissiveFactor": [7.0],
                    "fSpecularPower": [25.0],
                    "vAlbedoEmissiveColorMultiplier": [7.0, 0.15, 0.0, 1.0],
                    "vConstant0": [1.0, 1.0, 1.0, 1.0]
                },
                sampler_params={
                    "texBaseColor": {"filepath": ("actors/props/doorshieldstormmissile/models"
                                                  "/textures/doorshieldstormmissile_alt_bc.bctex")},
                    "texAttributes": {"filepath": SMOOTH_ATTRIBUTES},
                    "texNormals": {"filepath": SMOOTH_NORMALS}
                }
            )
        ]
    ),

    "bomb": ShieldData(
        name="doorshieldbomb",
        type=DoorTemplates.TRIANGLES,

        # "BOMB" weakness is added in runtime lua.
        # see open_dread_rando/files/custom_scenario.lua : Scenario._BlastShieldTypes
        weaknesses=[],

        collision="actors/props/doorshieldmissile/collisions/shield_bomb_colls.bmscd",
        actordef="actors/props/doorshieldbomb/charclasses/doorshieldbomb.bmsad",
        default_mdl=ModelData(
            base_model=SUPER_MDL,
            new_path="actors/props/doorshieldbomb/models/doorshieldbomb.bcmdl",
            materials={
                "mp_opaque_01": "actors/props/doorshieldbomb/models/imats/doorshieldbomb_mp_opaque_01.bsmat"
            }
        ),
        default_mats=[
            MaterialData(
                base_mat=SUPER_MAT,
                new_mat_name="doorshieldbomb_mp_opaque_01",
                new_path="actors/props/doorshieldbomb/models/imats/doorshieldbomb_mp_opaque_01.bsmat",
                uniform_params={
                    "vAlbedoEmissiveColorMultiplier": [0.987, 0.168, 0.976, 1.0]
                },
                sampler_params={
                    "texBaseColor": {"filepath": ("actors/props/doorshieldbomb/models"
                                                  "/textures/doorshieldbomb_bc.bctex")},
                    "texAttributes": {"filepath": SMOOTH_ATTRIBUTES},
                    "texNormals": {"filepath": SMOOTH_NORMALS}
                }
            )
        ],
        alternate_mats=[
            MaterialData(
                base_mat=SUPER_MAT,
                new_mat_name="doorshieldbomb_mp_opaque_01",
                new_path="actors/props/doorshieldbomb/models/imats/doorshieldbomb_mp_opaque_01.bsmat",
                uniform_params={
                    "vAlbedoEmissiveColorMultiplier": [0.987, 0.168, 0.976, 1.0]
                },
                sampler_params={
                    "texBaseColor": {"filepath": ("actors/props/doorshieldbomb/models"
                                                  "/textures/doorshieldbomb_alt_bc.bctex")},
                    "texAttributes": {"filepath": SMOOTH_ATTRIBUTES},
                    "texNormals": {"filepath": SMOOTH_NORMALS}
                }
            )
        ]
    ),

    "cross_bomb": ShieldData(
        name="doorshieldcrossbomb",
        type=DoorTemplates.TRIANGLES,
        weaknesses=["LINE_BOMB"],
        collision="actors/props/doorshieldmissile/collisions/shield_bomb_colls.bmscd",
        actordef="actors/props/doorshieldcrossbomb/charclasses/doorshieldcrossbomb.bmsad",
        default_mdl=ModelData(
            base_model=SUPER_MDL,
            new_path="actors/props/doorshieldcrossbomb/models/doorshieldcrossbomb.bcmdl",
            materials={
                "mp_opaque_01": ("actors/props/doorshieldcrossbomb/models"
                                 "/imats/doorshieldcrossbomb_mp_opaque_01.bsmat")
            }
        ),
        default_mats=[
            MaterialData(
                base_mat=SUPER_MAT,
                new_mat_name="doorshieldcrossbomb_mp_opaque_01",
                new_path="actors/props/doorshieldcrossbomb/models/imats/doorshieldcrossbomb_mp_opaque_01.bsmat",
                uniform_params={
                    "vAlbedoEmissiveColorMultiplier": [0.987, 0.168, 0.976, 1.0]
                },
                sampler_params={
                    "texBaseColor": {"filepath": ("actors/props/doorshieldpowerbomb/models"
                                                  "/textures/doorshieldpowerbomb_rdv_bc.bctex")},
                    "texAttributes": {"filepath": SMOOTH_ATTRIBUTES},
                    "texNormals": {"filepath": SMOOTH_NORMALS}
                }
            )
        ],
        alternate_mats=[
            MaterialData(
                base_mat=SUPER_MAT,
                new_mat_name="doorshieldcrossbomb_mp_opaque_01",
                new_path="actors/props/doorshieldcrossbomb/models/imats/doorshieldcrossbomb_mp_opaque_01.bsmat",
                uniform_params={
                    "vAlbedoEmissiveColorMultiplier": [0.987, 0.168, 0.976, 1.0]
                },
                sampler_params={
                    # TODO cross bomb basecolor
                    "texBaseColor": {"filepath": ("actors/props/doorshieldcrossbomb/models"
                                                  "/textures/doorshieldcrossbomb_alt_bc.bctex")},
                    "texAttributes": {"filepath": SMOOTH_ATTRIBUTES},
                    "texNormals": {"filepath": SMOOTH_NORMALS}
                }
            )
        ]
    ),

    "power_bomb": ShieldData(
        name="doorshieldpowerbomb",
        type=DoorTemplates.TRIANGLES,
        weaknesses=["POWER_BOMB"],
        collision="actors/props/doorshieldmissile/collisions/shield_bomb_colls.bmscd",
        actordef="actors/props/doorshieldpowerbomb/charclasses/doorshieldpowerbomb.bmsad",
        default_mdl=ModelData(
            base_model=SUPER_MDL,
            new_path="actors/props/doorshieldpowerbomb/models/doorshieldpowerbomb.bcmdl",
            materials={
                "mp_opaque_01": ("actors/props/doorshieldpowerbomb/models"
                                 "/imats/doorshieldpowerbomb_mp_opaque_01.bsmat")
            }
        ),
        default_mats=[
            MaterialData(
                base_mat=SUPER_MAT,
                new_mat_name="doorshieldpowerbomb_mp_opaque_01",
                new_path="actors/props/doorshieldpowerbomb/models/imats/doorshieldpowerbomb_mp_opaque_01.bsmat",
                uniform_params={
                    "vAlbedoEmissiveColorMultiplier": [1.0, 0.15, 0.0, 1.0]
                },
                sampler_params={
                    "texBaseColor": {"filepath": ("actors/props/doorshieldpowerbomb/models"
                                                  "/textures/doorshieldpowerbomb_rdv_bc.bctex")},
                    "texAttributes": {"filepath": SMOOTH_ATTRIBUTES},
                    "texNormals": {"filepath": SMOOTH_NORMALS}
                }
            )
        ],
        alternate_mats=[
            MaterialData(
                base_mat=SUPER_MAT,
                new_mat_name="doorshieldpowerbomb_mp_opaque_01",
                new_path="actors/props/doorshieldpowerbomb/models/imats/doorshieldpowerbomb_mp_opaque_01.bsmat",
                uniform_params={
                    "vAlbedoEmissiveColorMultiplier": [1.0, 0.15, 0.0, 1.0]
                },
                sampler_params={
                    "texBaseColor": {"filepath": ("actors/props/doorshieldpowerbomb/models"
                                                  "/textures/doorshieldpowerbomb_alt_bc.bctex")},
                    "texAttributes": {"filepath": SMOOTH_NORMALS},
                    "texNormals": {"filepath": SMOOTH_NORMALS}
                }
            )
        ],
    ),

    "closed": ShieldData(
        name="doorshieldclosed",
        type=DoorTemplates.TRIANGLES,
        weaknesses=[],
        actordef="actors/props/doorshieldclosed/charclasses/doorshieldclosed.bmsad",
        default_mdl=ModelData(
            base_model=SUPER_MDL,
            new_path="actors/props/doorshieldclosed/models/doorshieldclosed.bcmdl",
            materials={
                "mp_opaque_01": ("actors/props/doorshieldclosed/models"
                                 "/imats/doorshieldclosed_mp_opaque_01.bsmat")
            }
        ),
        default_mats=[
            MaterialData(
                base_mat=SUPER_MAT,
                new_mat_name="doorshieldclosed_mp_opaque_01",
                new_path="actors/props/doorshieldclosed/models/imats/doorshieldclosed_mp_opaque_01.bsmat",
                uniform_params={
                    "vAlbedoEmissiveColorMultiplier": [1.0, 0.0, 0.0, 1.0]
                },
                sampler_params={
                    "texBaseColor": {"filepath": ("actors/props/doorshieldclosed/models"
                                                  "/textures/doorshieldclosed_bc.bctex")},
                    "texAttributes": {"filepath": SMOOTH_ATTRIBUTES},
                    "texNormals": {"filepath": SMOOTH_NORMALS}
                }
            )
        ],
        alternate_mats=[
            MaterialData(
                base_mat=SUPER_MAT,
                new_mat_name="doorshieldclosed_mp_opaque_01",
                new_path="actors/props/doorshieldclosed/models/imats/doorshieldclosed_mp_opaque_01.bsmat",
                uniform_params={
                    "vAlbedoEmissiveColorMultiplier": [1.0, 0.0, 0.0, 1.0]
                },
                sampler_params={
                    "texBaseColor": {"filepath": ("actors/props/doorshieldclosed/models"
                                                  "/textures/doorshieldclosed_alt_bc.bctex")},
                    "texAttributes": {"filepath": SMOOTH_ATTRIBUTES},
                    "texNormals": {"filepath": SMOOTH_NORMALS}
                }
            )
        ]
    )
}


class BaseShield:
    def __init__(self, shield: ShieldData):
        self.data = shield

    def patch_model_data(self, new_template: Bmsad, editor: PatcherEditor, version: str) -> None:
        # select the default or alternate model
        if version == "DEFAULT" or self.data.alternate_mdl is None:
            model = self.data.default_mdl
        else:
            model = self.data.alternate_mdl

        create_custom_model(editor, model)
        assert model.new_path is not None
        new_template.model_name = model.new_path
        model_updater = new_template.components["MODELUPDATER"]
        model_updater.functions[0].set_param(1, model.new_path)

    def patch_material_data(self, editor: PatcherEditor, version: str) -> None:
        if version == "DEFAULT":
            for mat_dat in self.data.default_mats:
                create_custom_material(editor, mat_dat)

        else:
            for mat_dat in self.data.alternate_mats:
                create_custom_material(editor, mat_dat)

    def add_weakness(self, weakness: str, new_template: Bmsad) -> None:
        life = new_template.components["LIFE"]

        new_func = ActorDefFunc.new("AddDamageSource")
        new_func.set_param(1, weakness)

        life_funcs = list(life.functions)
        life_funcs.append(new_func)
        life.functions = life_funcs

    def change_collision(self, new_template: Bmsad) -> None:
        if self.data.collision is None:
            return

        coll_comp = new_template.components["COLLISION"]
        assert isinstance(coll_comp.dependencies, dict)
        coll_comp.dependencies["file"] = self.data.collision

    def patch(self, editor: PatcherEditor, version: str = "DEFAULT") -> None:
        template_bmsad = _template_read_shield(self.data.type.value)

        new_template = Bmsad.parse(template_bmsad, editor.target_game)
        new_template.name = self.data.name

        self.patch_material_data(editor, version)
        self.patch_model_data(new_template, editor, version)

        for w in self.data.weaknesses:
            self.add_weakness(w, new_template)

        self.change_collision(new_template)

        editor.add_new_asset(self.data.actordef, new_template, [])


def create_all_shield_assets(editor: PatcherEditor, shield_model_config: dict[str, str]) -> None:
    for shield_name, shield_type in shield_model_config.items():
        BaseShield(ALL_SHIELD_DATA[shield_name]).patch(editor, shield_type)



import shutil
from pathlib import Path

from open_dread_rando.files import files_path


def patch_exefs(exefs_patches: Path, configuration: dict):
    exefs_patches.mkdir(parents=True, exist_ok=True)

    provided_patches = files_path().joinpath("exefs_patches")
    shutil.copytree(provided_patches, exefs_patches, dirs_exist_ok=True)


def include_depackager(exefs_path: Path):
    exefs_path.mkdir(parents=True, exist_ok=True)

    dread_depackager = files_path().joinpath("dread_depackager")
    shutil.copytree(dread_depackager, exefs_path, dirs_exist_ok=True)



from enum import Enum

# constant data, including "magic numbers" and common data that many modules use

# list of all scenarios
ALL_SCENARIOS = [
    "s010_cave", "s020_magma",
    "s030_baselab",
    "s040_aqua",
    "s050_forest",
    "s060_quarantine",
    "s070_basesanc",
    "s080_shipyard",
    "s090_skybase"
]

# fade in/out values (in seconds) for room name GUI
class FadeTimes(Enum):
    NO_FADE = -1
    ROOM_FADE = 3



from math import ceil

from open_dread_rando.patcher_editor import PatcherEditor
from open_dread_rando.specific_patches.environmental_damage_sources import ALL_DAMAGE_ROOM_ACTORS


def get_damage_and_tick(value: float) -> tuple[float, float]:
    if value <= 0:
        damage = 0  # Damage value of 0 to bypass divide-by-zero later
        tick = 0.5  # Tick length of 0.5 is default for damage rooms, lava uses 1 second, but this should not matter.
    else:
        # Damage per tick (or 'fDamagePerTime') must be a whole number, this calculation
        # finds a value of damage that will be multiplied by the tick length to equal the
        # total DPS, such that the tick length is between 0.2s and 0.4s
        damage = ceil(value / 5)

        # Tick length (or 'fInBetweenDamageTime') is found by dividing the damage value by the DPS input,
        # and a value of 1.5/60 is subtracted from the result.
        # 1/60 is meant to account for the frame that damage is applied to Samus,
        # which does not count toward the timer, but 1.5/60 worked more accurately in testing for an unknown reason.
        tick = damage / value - (1.5 / 60)

    return damage, tick


def apply_constant_damage(editor: PatcherEditor, configuration: dict):
    for reference in ALL_DAMAGE_ROOM_ACTORS:
        actor = editor.resolve_actor_reference(reference)

        if actor.oActorDefLink == "actordef:actors/props/env_frozen_gen_001/charclasses/env_frozen_gen_001.bmsad":
            config_name = 'oFreezeConfig'
            config_field = "cold"
        elif actor.oActorDefLink == "actordef:actors/props/env_heat_gen_001/charclasses/env_heat_gen_001.bmsad":
            config_name = 'oHeatConfig'
            config_field = "heat"
        elif actor.oActorDefLink == "actordef:actors/props/lavazone/charclasses/lavazone.bmsad":
            config_name = 'oConfig'
            config_field = "lava"
        else:
            raise ValueError(f"{reference} does not have a valid actorDef for environmental damage")

        if configuration[config_field] is None:
            continue

        damage, tick = get_damage_and_tick(configuration[config_field])

        damage_config = actor.pComponents.ACTIVATABLE[config_name]
        damage_config.fDamagePerTime = damage
        damage_config.fInBetweenDamageTime = tick
        # Setting this value to 0 removes the damage ramp that exists in vanilla for damage rooms.
        damage_config.fDamageIncreaseAmount = 0
        # 1000 Max damage is an arbitrary limit, also set as the maximum in Randovania.
        # There is no reason this could not be larger if needed.
        damage_config.fMaxDamage = 1000



import dataclasses
import typing

from mercury_engine_data_structures.formats.bcmdl import Bcmdl

from open_dread_rando.misc_patches.material_patcher import MaterialData, create_custom_material
from open_dread_rando.patcher_editor import PatcherEditor


@dataclasses.dataclass(frozen=True)
class ModelData:
    base_model: str  # the model this is based on

    # the new path of the model
    new_path: typing.Optional[str] = None

    # dictionary connecting model names (i.e. "mp_opaque_01") to material files
    materials: typing.Optional[dict[str, str]] = None

# contains static data to generate model/material changes
@dataclasses.dataclass()
class StaticModelChanger:
    base_model_path: str
    new_model_path: str
    materials: dict[str, MaterialData] # key is material name in bcmdl

    def generate(self, editor: PatcherEditor):
        mats: dict[str, str] = {}

        for name, mat in self.materials.items():
            create_custom_material(editor, mat)
            mats[name] = mat.new_path

        create_custom_model(editor, ModelData(
            base_model=self.base_model_path,
            new_path=self.new_model_path,
            materials=mats
        ))


ADVANCED_RECOLORS: dict[str, StaticModelChanger] = {
    "item_supermissiletank": StaticModelChanger(
        base_model_path="actors/items/item_missiletankplus/models/item_missiletankplus.bcmdl",
        new_model_path="actors/items/item_missiletankplus/models/super_missile_tank.bcmdl",
        materials={
            "mat01": MaterialData(
                base_mat="actors/items/item_missiletankplus/models/imats/item_missiletankplus_mat01.bsmat",
                new_mat_name="supermissiletank_mat01",
                new_path="actors/items/item_missiletankplus/models/imats/super_missile_tank_mat01.bsmat",
                uniform_params={
                    "vConstant0": (0.0, 10.0, 0.15, 1.0)
                }
            ),
            "mp_fxhologram_01": MaterialData(
                base_mat="actors/items/item_missiletankplus/models/imats/item_missiletankplus_mp_fxhologram_01.bsmat",
                new_mat_name="supermissiletank_mp_fxhologram_01",
                new_path="actors/items/item_missiletankplus/models/imats/super_missile_tank_mp_fxhologram_01.bsmat",
                uniform_params={
                    "vTex0EmissiveColor": (10.0, 10.0, 10.0, 1.0)
                }
            )
        }
    )
}

def create_custom_model(editor: PatcherEditor, model_data: ModelData) -> Bcmdl:
    mdl = editor.get_parsed_asset(model_data.base_model, type_hint=Bcmdl)

    if model_data.materials is not None:
        for mat_name, mat_path in model_data.materials.items():
            editor.check_file_exists(mat_path, "Custom model material")
            mdl.change_material_path(mat_name, mat_path)

    if model_data.new_path:
        editor.add_new_asset(model_data.new_path, mdl, [])

    return mdl

def generate_custom_models(editor: PatcherEditor):
    for _, recolor in ADVANCED_RECOLORS.items():
        recolor.generate(editor)



import dataclasses
from enum import Enum

from open_dread_rando.logger import LOG
from open_dread_rando.patcher_editor import PatcherEditor
from open_dread_rando.pickups.map_icons import MapIcon


class TransporterType(Enum):
    TRANSPORT = "TRANSPORT"
    TELEPORTAL = "TELEPORTAL"

@dataclasses.dataclass
class TransporterIcon:
    default_icon_id: str
    coords: tuple[int, int]
    prefix: str
    disabled_id: str = ''

    def map_icon(self, icon_id: str, label: str) -> MapIcon:
        return MapIcon(
            icon_id=icon_id,
            coords=self.coords,
            label=self.prefix + label.upper(),
            disabled_id=self.disabled_id
        )

    @classmethod
    def get_icon(cls, usable: dict):
        usable_id = usable["sIconId"]
        for trans_type in TRANSPORT_TYPES.values():
            if trans_type.default_icon_id == usable_id:
                return trans_type
        return None

TRANSPORT_TYPES = {
    "Elevator": TransporterIcon(
        default_icon_id="UsableElevator",
        coords=(2,4),
        prefix="ELEVATOR TO ",
        disabled_id="DisabledElevator",
    ),
    "Train": TransporterIcon(
        default_icon_id="UsableTrain",
        coords=(1,4),
        prefix="SHUTTLE TO ",
    ),
    "Transport": TransporterIcon(
        default_icon_id="UsableTransport",
        coords=(8,2),
        prefix = "TRANSPORT CAPSULE TO ",
    )
}

def _get_type_and_usable(editor: PatcherEditor, elevator: dict) -> tuple[TransporterType, dict]:
    scenario = editor.get_scenario(elevator["teleporter"]["scenario"])
    sublayer = elevator["teleporter"].get("sublayer", elevator["teleporter"].get("layer", "default"))
    actor = scenario.actors_for_sublayer(sublayer)[elevator["teleporter"]["actor"]]

    try:
        usable = actor.pComponents.USABLE
    except AttributeError:
        raise ValueError(f'Actor {elevator["teleporter"]} is not usable')

    if usable["@type"] in ["CElevatorUsableComponent", "CTrainUsableComponent", "CTrainUsableComponentCutScene",
                           "CTrainWithPortalUsableComponent", "CCapsuleUsableComponent"]:
        return TransporterType.TRANSPORT, usable
    elif usable["@type"] == "CTeleporterUsableComponent":
        return TransporterType.TELEPORTAL, usable
    else:
        raise ValueError(f"Elevator {elevator['teleporter']['scenario']}/{elevator['teleporter']['actor']} "
                         "is not an elevator, shuttle, capsule or teleporter!\n"
                         f"USABLE type: {usable['@type']}")

def _patch_actor(usable: dict, elevator: dict):
    usable.sScenarioName = elevator["destination"]["scenario"]
    usable.sTargetSpawnPoint = elevator["destination"]["actor"]

def _patch_minimap_arrows(editor: PatcherEditor, elevator: dict):
    map = editor.get_scenario_map(elevator["teleporter"]["scenario"])
    sign = map.get_category("mapTransportSigns")[ elevator["teleporter"]["actor"] ]
    sign["sDestAreaId"] = elevator["destination"]["scenario"]

def _patch_map_icon(editor: PatcherEditor, elevator: dict):
    # get transporter type from mapUsables entry
    conn_name: str = elevator["connection_name"]
    icon_id = f'RDV_TRANSPORT_{conn_name.replace(" ", "")}' # ex: "RDV_TRANSPORT_Artaria-ThermalDevice"

    bmmap = editor.get_scenario_map(elevator["teleporter"]["scenario"])
    usable = bmmap.get_category("mapUsables")[ elevator["teleporter"]["actor"] ]
    ti = TransporterIcon.get_icon(usable)
    if ti is None:
        raise ValueError(f'Usable icon_id {usable["sIconId"]} invalid!')

    # add BMMDEF
    editor.map_icon_editor.add_icon(ti.map_icon(icon_id, elevator["connection_name"]))

    # update usable
    usable["sIconId"] = icon_id

def patch_elevators(editor: PatcherEditor, elevators_config: list[dict]):
    for elevator in elevators_config:
        LOG.debug("Writing elevator from: %s", str(elevator["teleporter"]))
        transporter_type, usable = _get_type_and_usable(editor, elevator)

        if transporter_type == TransporterType.TRANSPORT:
            _patch_actor(usable, elevator)
            _patch_minimap_arrows(editor, elevator)
            _patch_map_icon(editor, elevator)
        else:
            # TODO implement teleporter rando
            _patch_actor(usable, elevator)



import copy
from typing import Optional

import construct
from mercury_engine_data_structures.formats import Bmmap, Brfld
from mercury_engine_data_structures.formats.dread_types import CDoorLifeComponent_SState
from mercury_engine_data_structures.formats.gui_files import Bmscp

from open_dread_rando.constants import ALL_SCENARIOS
from open_dread_rando.door_locks import door_patcher
from open_dread_rando.logger import LOG
from open_dread_rando.patcher_editor import PatcherEditor
from open_dread_rando.pickups.map_icons import MapIconEditor


def flip_icon_id(icon_id: str) -> str:
    if icon_id.endswith("R"):
        return icon_id[:-1] + "L"
    elif icon_id.endswith("L"):
        return icon_id[:-1] + "R"
    raise ValueError(f"Unable to flip icon {icon_id}")


def _apply_one_sided_door_fix(editor: PatcherEditor, scenario: Brfld, map_blockages: construct.Container,
                              sublayer_name: str, actor_name: str, actor: construct.Container):
    # Continue if this isn't a door
    if not door_patcher.is_door(actor):
        return

    life_comp = actor.pComponents.LIFE
    left = scenario.follow_link(life_comp.wpLeftDoorShieldEntity)
    right = scenario.follow_link(life_comp.wpRightDoorShieldEntity)

    if left is None and right is None:
        return

    elif left is None:
        other = right
        direction = "wpLeftDoorShieldEntity"

    elif right is None:
        other = left
        direction = "wpRightDoorShieldEntity"
    else:
        return

    if "db_hdoor" in other.oActorDefLink:
        return

    LOG.debug("%s/%s: copy %s into %s", sublayer_name, actor_name, other.sName, direction)
    mirrored = copy.deepcopy(other)
    mirrored.sName += "_mirrored"
    mirrored.vAng = [other.vAng[0], -other.vAng[1], other.vAng[2]]
    scenario.actors_for_sublayer(sublayer_name)[mirrored.sName] = mirrored

    mirrored_map = copy.deepcopy(map_blockages[other.sName])
    mirrored_map["sIconId"] = flip_icon_id(mirrored_map["sIconId"])
    delta = 450 if mirrored_map["sIconId"].endswith("R") else -450
    mirrored_map["oBox"]["Min"][0] += delta
    mirrored_map["oBox"]["Max"][0] += delta
    map_blockages[mirrored.sName] = mirrored_map

    # Add a reference to the other shield to the main actor
    life_comp[direction] = f"Root:pScenario:rEntitiesLayer:dctSublayers:{sublayer_name}:dctActors:{mirrored.sName}"

    for x in [other.sName, mirrored.sName]:
        editor.copy_actor_groups(
            { "actor": actor_name },
            { "actor": x },
            scenario.name,
        )


def apply_one_sided_door_fixes(editor: PatcherEditor):
    for scenario_name in ALL_SCENARIOS:
        scenario = editor.get_scenario(scenario_name)
        bmmap = editor.get_scenario_map(scenario_name)
        map_blockages = bmmap.raw.Root.mapBlockages
        actordef = "actors/props/doorpowerpower/charclasses/doorpowerpower.bmsad"

        for sublayer_name, actor_name, actor in editor.find_type_of_actor(scenario_name, actordef):
            _apply_one_sided_door_fix(editor, scenario, map_blockages, sublayer_name, actor_name, actor)


PROBLEM_LAYERS = {
    "PostXRelease": {
        "s010_cave": [
            "collision_camera_026",  # Chain reaction
            "collision_camera_020",  # Corpius arena
            "collision_camera_073",  # Corpius entrance
        ],
        "s020_magma": [
            "collision_camera_063",  # Kraid arena
        ],
        "s040_aqua": [
            "collision_camera_007",  # Below Drogyga
            "collision_camera_028",  # Drogyga arena
        ],
        "s070_basesanc": [
            "collision_camera_005",  # Quiet Robe room
        ]
    },
    "Cooldown": {
        "s020_magma": [
            "collision_camera_004",  # Z-57 Access
        ]
    },
    "PostEmmy": {
        "s070_basesanc": [
            "collision_camera_040",  # Purple Emmi Introduction
        ]
    }
}


def remove_problematic_x_layers(editor: PatcherEditor):
    # these X layers have priority, so they will set some rooms to their post-state
    # even if they've never been entered. in these particular problem rooms, this
    # can cause softlocks (e.g. phantom cloak on golzuna, main PBs on corpius)
    for setup, levels in PROBLEM_LAYERS.items():
        for level, layers in levels.items():
            manager = editor.get_subarea_manager(level)
            configs = manager.get_subarea_setup(setup).vSubareaConfigs
            manager.get_subarea_setup(setup).vSubareaConfigs = [
                config for config in configs if config.sId not in layers
            ]


def _apply_boss_cutscene_fixes(editor: PatcherEditor, cutscene_ref: dict, callback: str,
                               insert_callback_at: Optional[int] = None):
    cutscene_player = editor.resolve_actor_reference(cutscene_ref)
    callbacks_after_cutscene = cutscene_player.pComponents.CUTSCENE.vctOnAfterCutsceneEndsLA

    # Add the custom call when the boss dies
    if callback:
        callback_action = {
            "@type": "CLuaCallsLogicAction",
            "sCallbackEntityName": "",
            "sCallback": callback,
            "bCallbackEntity": False,
            "bCallbackPersistent": False,
        }
        if insert_callback_at is None:
            callbacks_after_cutscene.append(callback_action)
        else:
            i = insert_callback_at
            callbacks_after_cutscene[i:i] = [callback_action]


def apply_corpius_fixes(editor: PatcherEditor):
    _apply_boss_cutscene_fixes(editor, {
        "scenario": "s010_cave",
        "sublayer": "Cutscenes",
        "actor": "cutsceneplayer_57"
    }, "CurrentScenario.OnCorpiusDeath_CUSTOM", 0)


def apply_kraid_fixes(editor: PatcherEditor):
    _apply_boss_cutscene_fixes(editor, {
        "scenario": "s020_magma",
        "sublayer": "cutscenes",
        "actor": "cutsceneplayer_61"
    }, "CurrentScenario.OnKraidDeath_CUSTOM", -1)


def apply_drogyga_fixes(editor: PatcherEditor):
    _apply_boss_cutscene_fixes(editor, {
        "scenario": "s040_aqua",
        "sublayer": "cutscenes",
        "actor": "cutsceneplayer_65"
    }, "CurrentScenario.OnHydrogigaDead_CUSTOM", -1)

    # remove the trigger that deletes drogyga until after beating drogyga
    aqua = editor.get_scenario("s040_aqua")
    aqua.remove_actor_from_group("eg_collision_camera_007_Default", "TG_WaterPoolAfterHydrogiga", "Boss")


def activate_emmi_zones(editor: PatcherEditor):
    # Remove the cutscene that plays when you enter the emmi zone for the first time
    # This causes the border of the zone to not be revealed, but it should be possible to do that in another way
    editor.remove_entity(
        {
            "scenario": "s010_cave",
            "sublayer": "Cutscenes",
            "actor": "cutscenetrigger_36"
        },
        None,
    )

    # Remove the cutscene that introduces the speed emmi
    # This causes the border of the zone to not be revealed, but it should be possible to do that in another way
    editor.remove_entity(
        {
            "scenario": "s030_baselab",
            "sublayer": "cutscenes",
            "actor": "cutscenetrigger_39"
        },
        None,
    )


def fix_backdoor_white_cu(editor: PatcherEditor):
    new_door = construct.Container({
        "@type": "CEntity",
        "sName": "DreadRando_CUDoor",
        "oActorDefLink": "actordef:actors/props/doorclosedpower/charclasses/doorclosedpower.bmsad",
        "vPos": [
            8300.000,
            -4700.000,
            0.0
        ],
        "vAng": [
            0.0,
            0.0,
            0.0
        ],
        "pComponents": {
            "LIFE": {
                "@type": "CDoorLifeComponent",
                "bWantsEnabled": True,
                "bUseDefaultValues": True,
                "fMaxDistanceOpened": 400.0,
                "wpLeftDoorShieldEntity": "{EMPTY}",
                "wpRightDoorShieldEntity": "{EMPTY}",
                "fMinTimeOpened": 3.0,
                "bStayOpen": True,
                "bStartOpened": False,
                "bOnBlackOutOpened": False,
                "bDoorIsWet": False,
                "bFrozenDuringColdown": True,
                "iAreaLeft": 0,
                "iAreaRight": 0,
                "aVignettes": []
            },
            "AUDIO": {
                "@type": "CAudioComponent",
                "bWantsEnabled": True,
                "bUseDefaultValues": True
            },
            "FX": {
                "@type": "CFXComponent",
                "bWantsEnabled": True,
                "bUseDefaultValues": True,
                "fSelectedHighRadius": 250.0,
                "fSelectedLowRadius": 350.0
            },
            "COLLISION": {
                "@type": "CCollisionComponent",
                "bWantsEnabled": True,
                "bUseDefaultValues": True
            },
            "TIMELINECOMPONENT": {
                "@type": "CTimelineComponent",
                "bWantsEnabled": True,
                "bUseDefaultValues": True
            },
            "NAVMESHITEM": {
                "@type": "CNavMeshItemComponent",
                "bWantsEnabled": True,
                "bUseDefaultValues": True,
                "tForbiddenEdgesSpawnPoints": []
            },
            "ANIMATION": {
                "@type": "CAnimationComponent",
                "bWantsEnabled": True,
                "bUseDefaultValues": True
            },
            "MODELUPDATER": {
                "@type": "CModelUpdaterComponent",
                "bWantsEnabled": True,
                "bUseDefaultValues": True,
                "sDefaultModelPath": ""
            }
        },
        "bEnabled": True
    })

    cave = editor.get_scenario("s010_cave")
    cave.actors_for_sublayer("default")[new_door["sName"]] = new_door

    for group in ["eg_collision_camera_018_Default", "eg_collision_camera_090_Default",
                  "eg_collision_camera_049_Default", "eg_collision_camera_090_PostXRelease",
                  "eg_collision_camera_049_PostXRelease"]:
        cave.add_actor_to_group(group, new_door["sName"])


def patch_corpius_checkpoints(editor: PatcherEditor):
    cave = editor.get_scenario("s010_cave")

    # Fix checkpoint before Corpius fight not working if the X were released
    cave.add_actor_to_group("eg_collision_camera_072_PostXRelease", "SP_Checkpoint_Scorpius")


def apply_experiment_fixes(editor: PatcherEditor):
    magma = editor.get_scenario("s020_magma")
    magma_map = editor.get_scenario_map("s020_magma")

    _apply_boss_cutscene_fixes(editor, {
        "scenario": "s020_magma",
        "sublayer": "cutscenes",
        "actor": "cutsceneplayer_81"
    }, "CurrentScenario.OnExperimentDeath_CUSTOM", 0)

    new_triggers = {
        "TriggerEnableCooldown": (5050.000, -5346.150, 0.000),
        "TriggerDisableCooldown": (5000.000, -7350.000, 0.000),
    }

    # create triggers to update the cooldown status appropriately
    for name, pos in new_triggers.items():
        ap_trigger = copy.deepcopy(editor.resolve_actor_reference({
            "scenario": "s020_magma",
            "actor": "AP_03"
        }))

        ap_trigger.sName = name
        ap_trigger.vPos = pos

        activation_conditions = ap_trigger.pComponents.TRIGGER.lstActivationConditions
        activation_conditions[0].vLogicActions[0].sCallback = f"CurrentScenario.OnEnter_{name}"

        magma.actors_for_sublayer('default')[name] = ap_trigger
        magma.add_actor_to_group("eg_collision_camera_004_PostXRelease", name)

    # make thermal doors always closed during the fight
    for name in ["trap_thermal_horizontal_000", "trap_thermal_horizontal_005"]:
        magma.remove_actor_from_group("eg_collision_camera_009_Cooldown", name)

        trap = copy.deepcopy(editor.resolve_actor_reference({
            "scenario": "s020_magma",
            "actor": name
        }))

        trap.sName = f"{name}_EXPERIMENT"
        magma.actors_for_sublayer('default')[trap.sName] = trap
        magma.add_actor_to_group('eg_collision_camera_009_Cooldown', trap.sName)

    # disable closing the thermal door permanently after experiment
    editor.remove_entity({
        "scenario": "s020_magma",
        "actor": "trap_thermal_horizontal_POSTCOOL"
    }, "mapDoors")

    # add thermal door in front of morph launcher
    new_door = copy.deepcopy(editor.resolve_actor_reference({
        "scenario": "s020_magma",
        "actor": "trap_thermal_horizontal_004"
    }))

    new_name = "trap_thermal_horizontal_006"

    new_door.sName = new_name
    new_door.vPos = (5840.0, -5455.0, 0.0)

    magma.actors_for_sublayer('default')[new_name] = new_door
    editor.copy_actor_groups(
        {"actor": "trap_thermal_horizontal_004" },
        { "actor": new_name },
        "s020_magma"
    )

    # update the minimap
    new_map_icon = copy.deepcopy(magma_map.get_category("mapDoors")["trap_thermal_horizontal_004"])
    new_map_icon.vPos = new_door.vPos[:2]
    new_map_icon.oBoxL.Min = [c + offset for c, offset in zip(new_door.vPos, (-200.0, -50.0))]
    new_map_icon.oBoxL.Max = [c + offset for c, offset in zip(new_door.vPos, (200.0, 50.0))]
    magma_map.get_category("mapDoors")["trap_thermal_horizontal_006"] = new_map_icon

    # update thermal switch to open new door
    thermal_switch = editor.resolve_actor_reference({
        "scenario": "s020_magma",
        "actor": "deviceheat"
    })

    thermal_switch.pComponents.USABLE.vThermalDoors.append({
        "wpThermalDoor": magma.link_for_actor(new_name),
        "sDoorState": CDoorLifeComponent_SState.Opened
    })


def apply_main_menu_fixes(editor: PatcherEditor):
    extras = editor.get_file("gui/scripts/extrasmenucomposition.bmscp", Bmscp)
    listcomp = extras.get_child("Content.ListComposition").lstChildren
    listcomp.pop(2)  # remove the credits button from the extras menu


def disable_hanubia_cutscene(editor: PatcherEditor):
    # disable cutscene 12 (hanubia - tank room) because it teleports samus to the lower section (bad for door rando)
    cutscene_player = editor.resolve_actor_reference({
        "scenario": "s080_shipyard",
        "sublayer": "cutscenes",
        "actor": "cutsceneplayer_12",
    })
    cutscene_player.bEnabled = False


def fix_map_icons(map_editor: MapIconEditor):
    map_editor.mirror_bmmdef_icons()
    map_editor.mirror_bmmap_icons()


def move_artaria_missile_tank(editor: PatcherEditor):
    actor_ref = {
        "scenario": "s010_cave",
        "actor": "item_missiletank_000"
    }
    missile_tank = editor.resolve_actor_reference(actor_ref)
    artaria_map = editor.get_scenario_file(actor_ref["scenario"], Bmmap)

    new_x_pos = 15450.0
    missile_tank.vPos[0] = new_x_pos
    artaria_map.items[actor_ref["actor"]].vPos[0] = new_x_pos


def apply_static_fixes(editor: PatcherEditor):
    remove_problematic_x_layers(editor)
    activate_emmi_zones(editor)
    apply_one_sided_door_fixes(editor)
    apply_kraid_fixes(editor)
    apply_corpius_fixes(editor)
    fix_backdoor_white_cu(editor)
    patch_corpius_checkpoints(editor)
    apply_experiment_fixes(editor)
    apply_drogyga_fixes(editor)
    apply_main_menu_fixes(editor)
    disable_hanubia_cutscene(editor)
    fix_map_icons(editor.map_icon_editor)
    move_artaria_missile_tank(editor)



from unittest.mock import MagicMock

from open_dread_rando import dread_patcher


def test_cosmetic_options(lua_runtime):
    # Setup
    layoutUUID = "33aead28-ab1f-4d55-bbd9-fc0da5b48469"

    editor = MagicMock()
    configuration = {
        "starting_items": {},
        "starting_location": {
            "scenario": "s010_cave",
            "actor": "StartPoint0"
        },
        "configuration_identifier": "<identifier>",
        "layout_uuid": layoutUUID,
        "cosmetic_patches": {
            "config": {
            },
            "lua": {
                "custom_init": {
                    "enable_death_counter": False,
                    "enable_room_name_display": "NEVER"
                },
                "camera_names_dict": {
                }
            },
            "shield_versions": {
                "ice_missile": "ALTERNATE",
                "diffusion_beam": "ALTERNATE",
                "storm_missile": "ALTERNATE",
                "bomb": "ALTERNATE",
                "cross_bomb": "ALTERNATE",
                "power_bomb": "DEFAULT"
            }
        },
        "energy_per_tank": 75,
        "immediate_energy_parts": True,
        "constant_environment_damage": {
            "heat": 20,
            "cold": 20,
            "lava": 20
        },
        "game_patches": {
            "raven_beak_damage_table_handling": "consistent_low",
            "remove_grapple_blocks_hanubia_shortcut": True,
            "remove_grapple_block_path_to_itorash": True,
            "default_x_released": False,
            "enable_experiment_boss": True,
            "warp_to_start": True,
        },
        "objective": {
            "required_artifacts": 3,
            "hints": []
        },
    }

    # Run
    result = dread_patcher.create_custom_init(editor, configuration)

    # Assert
    lua_runtime.execute("Init = {}")
    lua_runtime.execute(result)

    assert lua_runtime.eval("Init.fEnergyPerTank") == 75
    assert lua_runtime.eval("Init.sLayoutUUID") == layoutUUID



import copy
import shutil
import typing
from pathlib import Path

from construct import Container
from mercury_engine_data_structures.base_resource import BaseResource
from mercury_engine_data_structures.file_tree_editor import FileTreeEditor
from mercury_engine_data_structures.formats import ALL_FORMATS, Bmmap, Brfld, Brsa
from mercury_engine_data_structures.formats.brfld import ActorLayer
from mercury_engine_data_structures.game_check import Game
from mercury_engine_data_structures.romfs import ExtractedRomFs

from open_dread_rando.pickups.map_icons import MapIconEditor

T = typing.TypeVar("T")


def path_for_level(level_name: str) -> str:
    return f"maps/levels/c10_samus/{level_name}/{level_name}"


def extension_for_type(type_hint: type[T]) -> str:
    return next(ext for ext, t in ALL_FORMATS.items() if t == type_hint).lower()


class PatcherEditor(FileTreeEditor):
    memory_files: dict[str, BaseResource]
    map_icon_editor: MapIconEditor

    def __init__(self, root: Path):
        super().__init__(ExtractedRomFs(root), target_game=Game.DREAD)
        self.memory_files = {}
        self.map_icon_editor = MapIconEditor(self)

    def get_file(self, path: str, type_hint: type[T] = BaseResource) -> T:
        if path not in self.memory_files:
            self.memory_files[path] = self.get_parsed_asset(path, type_hint=type_hint)
        return self.memory_files[path]

    def get_scenario_file(self, name: str, type_hint: type[T]) -> T:
        path = f"{path_for_level(name)}.{extension_for_type(type_hint)}"
        return self.get_file(path, type_hint)

    def get_scenario(self, name: str) -> Brfld:
        return self.get_scenario_file(name, Brfld)

    def get_subarea_manager(self, name: str) -> Brsa:
        return self.get_scenario_file(name, Brsa)

    def get_scenario_map(self, name: str) -> Bmmap:
        return self.get_scenario_file(name, Bmmap)

    def get_level_pkgs(self, name: str) -> set[str]:
        return set(self.find_pkgs(path_for_level(name) + ".brfld"))

    def ensure_present_in_scenario(self, scenario: str, asset):
        for pkg in self.get_level_pkgs(scenario):
            self.ensure_present(pkg, asset)

    def resolve_actor_reference(self, ref: dict) -> Container:
        scenario = self.get_scenario(ref["scenario"])
        actor_layer = ActorLayer(ref.get("actor_layer", "rEntitiesLayer"))
        sublayer = ref.get("sublayer", ref.get("layer", "default"))
        return scenario.actors_for_sublayer(sublayer, actor_layer)[ref["actor"]]

    def flush_modified_assets(self):
        for name, resource in self.memory_files.items():
            self.replace_asset(name, resource)
        self.memory_files = {}

    def add_new_asset(self, name: str, new_data: typing.Union[bytes, BaseResource], in_pkgs: typing.Iterable[str]):
        super().add_new_asset(name, new_data, in_pkgs)
        # Hack for textures' weird folder layout
        if name.startswith("textures/") and isinstance(new_data, bytes):
            self._toc.add_file(name[9:], len(new_data))

    def save_modified_saves_to(self, debug_path: Path):
        shutil.rmtree(debug_path, ignore_errors=True)
        for asset_id, asset in self._modified_resources.items():
            if asset is not None:
                path = debug_path.joinpath(self._name_for_asset_id[asset_id])
                path.parent.mkdir(parents=True, exist_ok=True)
                path.write_bytes(asset)

    def remove_entity(self, reference: dict, map_category: typing.Optional[str]):
        scenario = self.get_scenario(reference["scenario"])
        actor_layer = ActorLayer(reference.get("actor_layer", "rEntitiesLayer"))
        sublayer = reference.get("sublayer", reference.get("layer", "default"))
        actor_name = reference["actor"]

        for group_name in scenario.actor_groups_for_actor_layer(actor_layer):
            if scenario.is_actor_in_group(group_name, actor_name, sublayer, actor_layer):
                scenario.remove_actor_from_group(group_name, actor_name, sublayer, actor_layer)

        scenario.actors_for_sublayer(sublayer, actor_layer).pop(actor_name)
        if map_category is not None:
            self.get_scenario_map(reference["scenario"]).raw.Root[map_category].pop(actor_name)

    def copy_actor_groups(self, base_reference: dict, new_reference: dict, scenario_name: str,
                          actor_layer: ActorLayer = ActorLayer.ENTITIES):
        """
        Copies a base actor's groups to a new actor's groups. Both actors must be in the same scenario and actor layer.

        param base_reference: a reference to the actor to copy the groups from
        param new_reference: a reference to the actor to copy the groups to
        """
        scenario = self.get_scenario(scenario_name)

        base_sublayer = base_reference.get("sublayer", base_reference.get("layer", "default"))
        base_actor_name = base_reference["actor"]
        new_sublayer = new_reference.get("sublayer", new_reference.get("layer", base_sublayer))
        new_actor_name = new_reference["actor"]

        for group_name in scenario.actor_groups_for_actor_layer(actor_layer):
            base_actor_in_group = scenario.is_actor_in_group(group_name, base_actor_name, base_sublayer, actor_layer)
            new_actor_in_group = scenario.is_actor_in_group(group_name, new_actor_name, new_sublayer, actor_layer)

            if base_actor_in_group and not new_actor_in_group:
                scenario.add_actor_to_group(group_name, new_actor_name, new_sublayer, actor_layer)
            elif not base_actor_in_group and new_actor_in_group:
                scenario.remove_actor_from_group(group_name, new_actor_name, new_sublayer, actor_layer)

    def copy_actor(self, scenario: str, coords, templateActor: Container, newName: str, offset: tuple = (0, 0, 0)):
        """
        Copies an actor into a scenario at a specific location

        @param scenario: the scenario to place the door
        @param coords: the position coordinate to place the door
        @param actor: The actor to be copied into door's scenario
        @param sName: The name of the new actor
        @param offset: a tuple containing the offset coordinates, if any
        @returns: the new copied actor
        """
        # print(coords)
        newActor = copy.deepcopy(templateActor)
        newActor.sName = newName
        currentScenario = self.get_scenario(scenario)
        currentScenario.actors_for_sublayer("default")[newActor.sName] = newActor
        newActor.vPos = [float(c) + offset for c, offset in zip(coords, offset)]

        return newActor

    def find_type_of_actor(self, scenario_name: str, actordef: str,
                           actor_layer: ActorLayer = ActorLayer.ENTITIES) -> list[tuple[str, str, Container]]:
        """
        Get every actor with given actordef in a scenario

        param scenario: the name of the scenario
        param actordef: the actor definition (bmsad) to filter for
        param sublayer: if provided, the sublayer to search in, otherwise default sublayer
        param actor_layer: if provided, the actor layer to search in, otherwise entities layer
        returns: for each actor that matches the criteria: actor name, actor
        """
        scenario = self.get_scenario(scenario_name)
        filtered = []

        for sublayer, actor_name, actor in scenario.all_actors_in_actor_layer(actor_layer):
            a = self.resolve_actor_reference({
                "actor": actor_name,
                "sublayer": sublayer,
                "actor_layer": actor_layer,
                "scenario": scenario_name
            })

            if a.oActorDefLink.split(':')[1] == actordef:
                filtered.append((sublayer, actor_name, actor))

        return filtered

    def reference_for_link(self, link: str, scenario: str) -> dict:
        """
        Changes a link string (wp data type) into a reference dict

        param link: a reference string consisting of 7 components separated by colons
        param scenario: the scenario this actor is in
        returns: a reference dict
        """
        split_link = link.split(":")
        if len(split_link) != 7:
            raise ValueError(f"Expected 7 components in {link}, got {len(split_link)}")

        actor_layer = split_link[2]
        sublayer = split_link[4]
        actor = split_link[6]

        return {
            "scenario": scenario,
            "actor_layer": actor_layer,
            "sublayer": sublayer,
            "actor": actor,
        }

    def build_link(self, sname: str, sublayer: str = "default", actor_layer: ActorLayer = ActorLayer.ENTITIES) -> str:
        return f"Root:pScenario:{actor_layer.value}:dctSublayers:{sublayer}:dctActors:{sname}"

    def get_asset_names_in_folder(self, folder: str) -> typing.Iterator[str]:
        yield from (name for name in self._name_for_asset_id.values() if name.startswith(folder))

    def check_file_exists(self, file_name: str, description: str = "Asset") -> None:
        actual_path = file_name
        if file_name.endswith(".bctex") and not file_name.startswith("textures/"):
            actual_path = "textures/" + file_name

        if not self.does_asset_exists(actual_path):
            raise ValueError(f"{description} {actual_path} does not exist.")



from mercury_engine_data_structures.formats import Bmsad

from open_dread_rando.misc_patches.text_patches import patch_text
from open_dread_rando.patcher_editor import PatcherEditor

_HANUBIA_SHORTCUT_GRAPPLE_BLOCKS = [
    {
        "scenario": "s080_shipyard",
        "actor": "grapplepulloff1x2_000"
    },
    {
        "scenario": "s080_shipyard",
        "actor": "grapplepulloff1x2"
    }
]


def apply_game_patches(editor: PatcherEditor, configuration: dict):
    raven_beak_damage_mode = configuration["raven_beak_damage_table_handling"]
    early_cloak_water_mode = configuration["remove_early_cloak_water"]
    arbitrary_enky_mode = configuration["remove_arbitrary_enky"]

    if raven_beak_damage_mode != "unmodified":
        _modify_raven_beak_damage_table(editor, raven_beak_damage_mode)

    _remove_grapple_blocks(editor, configuration)

    if configuration["warp_to_start"]:
        _warp_to_start(editor)

    if configuration["nerf_power_bombs"]:
        _remove_pb_weaknesses(editor)

    if configuration["remove_water_platform_water"]:
        _remove_water_platform_water(editor)

    if early_cloak_water_mode != "unmodified":
        _remove_early_cloak_water(editor, early_cloak_water_mode)

    if arbitrary_enky_mode != "unmodified":
        _remove_arbitrary_enky(editor, arbitrary_enky_mode)


def _modify_raven_beak_damage_table(editor: PatcherEditor, mode: str):
    rb_bmsad = editor.get_file("actors/characters/chozocommander/charclasses/chozocommander.bmsad", Bmsad)

    life_component = rb_bmsad.components["LIFE"]
    ai_component = rb_bmsad.components["AI"]

    if mode == "consistent_high":
        base_factor = life_component.fields.oDamageSourceFactor  # Base damage during regular fight
        counter_factors = [
            ai_component.fields.oDamageSourceFactorShortShootingGrab,  # Most counter cutscenes
            ai_component.fields.oDamageSourceFactorLongShootingGrab,  # Unknown alternate counter cutscene
        ]

        # Buffs Wave Beam and Ice Missiles to have the same damage VALUES (not factors) as vanilla Plasma Beam and
        #   Ice Missiles, respectively.
        # Wave Beam typically deals 1.6x the damage as Plasma Beam, so its relative factor must be 1/1.6 that of
        #   Plasma Beam (or 5/8, or 0.625)
        # Ice Missiles typically deal 1.33x the damage as Super Missiles, so their relative factor must be 1/1.33
        #   that of Super Missiles (or 3/4, or 0.75)

        # Base factors will yield the correct ratio such that wave and plasma deal identical damage (Ice uses standard)
        base_factor.fWaveBeamFactor = 0.625
        base_factor.fChargeWaveBeamFactor = 0.625
        base_factor.fMeleeChargeWaveBeamFactor = 0.625
        base_factor.fIceMissileFactor = 1.0

        # All melee-counter factors will be reset to 1.0, except Ice, which uses the factor explained above
        for factor in counter_factors:
            factor.fWaveBeamFactor = 1.0
            factor.fChargeWaveBeamFactor = 1.0
            factor.fMeleeChargeWaveBeamFactor = 1.0
            factor.fIceMissileFactor = 0.75
    else:
        # Debuffs all weapons prior to Wave Beam and Ice Missiles using the same damage factor as Wave Beam and
        # Ice Missiles have in vanilla
        factors = [
            life_component.fields.oDamageSourceFactor,  # Base damage during regular fight
            ai_component.fields.oDamageSourceFactorShortShootingGrab,  # Most counter cutscenes
            ai_component.fields.oDamageSourceFactorLongShootingGrab,  # Unknown alternate counter cutscene
        ]

        for factor in factors:
            for beam in ["fPowerBeamFactor", "fWideBeamFactor", "fPlasmaBeamFactor"]:
                factor[beam] = factor.fWaveBeamFactor
            for beam in ["fChargePowerBeamFactor", "fChargeWideBeamFactor", "fChargePlasmaBeamFactor"]:
                factor[beam] = factor.fChargeWaveBeamFactor
            for beam in ["fMeleeChargePowerBeamFactor", "fMeleeChargeWideBeamFactor", "fMeleeChargePlasmaBeamFactor"]:
                factor[beam] = factor.fMeleeChargeWaveBeamFactor
            for missile in ["fMissileFactor", "fSuperMissileFactor"]:
                factor[missile] = factor.fIceMissileFactor


def _remove_grapple_blocks(editor: PatcherEditor, configuration: dict):
    if configuration["remove_grapple_blocks_hanubia_shortcut"]:
        for reference in _HANUBIA_SHORTCUT_GRAPPLE_BLOCKS:
            editor.remove_entity(reference, "mapProps")

    if configuration["remove_grapple_block_path_to_itorash"]:
        editor.remove_entity(
            {
                "scenario": "s080_shipyard",
                "actor": "grapplepulloff1x2_001"
            },
            "mapProps"
        )


def _remove_pb_weaknesses(editor: PatcherEditor):
    # enky
    warlotus = editor.get_file("actors/characters/warlotus/charclasses/warlotus.bmsad", Bmsad)
    warlotus.components["LIFE"].fields.bShouldDieWithPowerBomb = False
    warlotus.components["LIFE"].fields.oDamageSourceFactor.fPowerBombFactor = 0.0

    # charge door
    for door in ["doorchargecharge", "doorchargeclosed", "doorclosedcharge"]:
        charge_door = editor.get_file(f"actors/props/{door}/charclasses/{door}.bmsad", Bmsad)
        func = charge_door.components["LIFE"].functions[0]

        if func.get_param(1):
            func.set_param(1, "CHARGE_BEAM")
        if func.get_param(2):
            func.set_param(2, "CHARGE_BEAM")


def _warp_to_start(editor: PatcherEditor):
    text = "Save your progress?|Hold \u1806 and \u1807 while selecting {c6}Cancel{c0}|to warp to the starting location."
    patch_text(editor, "GUI_SAVESTATION_QUESTION", text)


def _remove_water_platform_water(editor: PatcherEditor):
    editor.remove_entity(
        {
            "scenario": "s010_cave",
            "actor": "PRP_CV_watercave05"
        },
        "mapWaterPoolGeos"
    )

def _remove_early_cloak_water(editor: PatcherEditor, mode: str):
    editor.remove_entity(
        {
            "scenario": "s010_cave",
            "actor": "PRP_CV_watercave01b"
        },
        "mapWaterPoolGeos"
    )

    if mode == "both_sides":
        editor.remove_entity(
            {
                "scenario": "s010_cave",
                "actor": "PRP_CV_watercave01a"
            },
            "mapWaterPoolGeos"
        )
        editor.remove_entity(
            {
                "scenario": "s010_cave",
                "actor": "PRP_DB_CV_003"
            },
            "mapOccluderGeos"
        )


def _remove_arbitrary_enky(editor: PatcherEditor, mode: str):
    arbitrary_enky_reference = {
        "scenario": "s010_cave",
        "sublayer": "Enemies",
        "actor": "SG_WarLotus_000"
    }

    if mode == "never":
        editor.remove_entity(arbitrary_enky_reference, "mapProps")
    elif mode == "always":
        arbitrary_enky = editor.resolve_actor_reference(arbitrary_enky_reference)

        arbitrary_enky.pComponents.SPAWNGROUP.bAutoenabled = True



import itertools

from open_dread_rando.constants import ALL_SCENARIOS
from open_dread_rando.files import files_path
from open_dread_rando.misc_patches import lua_util
from open_dread_rando.patcher_editor import PatcherEditor, path_for_level


def _read_powerup_lua() -> bytes:
    return files_path().joinpath("randomizer_powerup.lua").read_bytes()


def _read_level_lua(level_id: str) -> str:
    return files_path().joinpath("levels", f"{level_id}.lc.lua").read_text()


SPECIFIC_CLASSES = {
    "ITEM_WEAPON_POWER_BOMB": "RandomizerPowerBomb",
    "ITEM_OPTIC_CAMOUFLAGE": "RandomizerPhantomCloak",
    "ITEM_SPEED_BOOSTER": "RandomizerSpeedBooster",
    "ITEM_MULTILOCKON": "RandomizerStormMissile",
    "ITEM_LIFE_SHARDS": "RandomizerEnergyPart",
    "ITEM_GHOST_AURA": "RandomizerFlashShift",
    "ITEM_WEAPON_POWER_BEAM": "RandomizerPowerBeam",
    "ITEM_WEAPON_WIDE_BEAM": "RandomizerWideBeam",
    "ITEM_WEAPON_PLASMA_BEAM": "RandomizerPlasmaBeam",
    "ITEM_WEAPON_WAVE_BEAM": "RandomizerWaveBeam",
    "ITEM_WEAPON_MISSILE_LAUNCHER": "RandomizerMissileLauncher",
    "ITEM_WEAPON_SUPER_MISSILE": "RandomizerSuperMissile",
    "ITEM_WEAPON_ICE_MISSILE": "RandomizerIceMissile",
}


class LuaEditor:
    def __init__(self):
        self._custom_classes = {}
        self._powerup_script = _read_powerup_lua()
        self._custom_level_scripts: dict[str, dict] = self._read_levels()
        self._corex_replacement = {
            "escue": "false",
            "golzuna": "false",
        }

    def _read_levels(self) -> dict[str, dict]:
        return {
            scenario: {"script": _read_level_lua(scenario), "edited": False}
            for scenario in ALL_SCENARIOS
        }

    def get_parent_for(self, item_id) -> str:
        return SPECIFIC_CLASSES.get(item_id, "RandomizerPowerup")

    def get_script_class(self, pickup: dict, boss: bool = False, actordef_name: str = "") -> str:
        pickup_resources = pickup["resources"]
        first_resource = pickup_resources[0][0]

        if not boss and len(pickup_resources) == 1 and len(pickup_resources[0]) == 1:
            first_resource_id = first_resource["item_id"]

            if "ITEM_RANDO_ARTIFACT_" in first_resource_id:
                if first_resource_id in self._custom_classes.keys():
                    return self._custom_classes[first_resource_id]

                class_name = f'RandomizerArtifact{first_resource_id[20:]}'

                self.add_custom_class(
                    {
                        "name": class_name,
                        "resources": [
                            [
                                {
                                    "item_id": lua_util.wrap_string(first_resource_id),
                                    "quantity": first_resource["quantity"]
                                }
                            ]
                        ],
                        "parent": "RandomizerPowerup"
                    }
                )

                self._custom_classes[first_resource_id] = class_name
                return class_name

            # Single-item pickup; don't include progressive template
            return self.get_parent_for(first_resource_id)

        if actordef_name and len(pickup["model"]) > 1:
            self.add_progressive_models(pickup, actordef_name)

        hashable_progression = "_".join([
            f'{res[0]["item_id"]}_{res[0]["quantity"]}'
            for res in pickup_resources
        ]).replace("-", "MINUS")

        if hashable_progression in self._custom_classes.keys():
            return self._custom_classes[hashable_progression]

        class_name = f"RandomizerProgressive{hashable_progression}"

        resources = [
            [
                {
                    "item_id": lua_util.wrap_string(res["item_id"]),
                    "quantity": res["quantity"]
                }
                for res in resource_list
            ]
            for resource_list in pickup_resources
        ]
        replacement = {
            "name": class_name,
            "resources": resources,
            "parent": self.get_parent_for(first_resource["item_id"]),
        }
        self.add_custom_class(replacement)

        self._custom_classes[hashable_progression] = class_name
        return class_name

    def add_custom_class(self, replacement):
        new_class = lua_util.replace_lua_template("custom_powerup_template.lua", replacement)
        self._powerup_script += new_class.encode("utf-8")

    def add_progressive_models(self, pickup: dict, actordef_name: str):
        progressive_models = [
            {
                "item": lua_util.wrap_string(resource["item_id"]),
                "alias": lua_util.wrap_string(model_name),
            }
            for resource, model_name in itertools.chain(
                zip([r[0] for r in pickup["resources"]], pickup["model"][1:]),
                [(pickup["resources"][-1][0], pickup["model"][-1])],
            )
        ]
        progressive_models.reverse()

        replacement = {
            "actordef_name": actordef_name,
            "progressive_models": progressive_models,
        }

        models = lua_util.replace_lua_template("progressive_model_template.lua", replacement)
        self._powerup_script += models.encode("utf-8")

    def patch_actordef_pickup_script(self, editor: PatcherEditor, pickup: dict, pickup_lua_callback: dict,
                                     extra_code: str = "") -> None:
        scenario = pickup_lua_callback["scenario"]
        scenario_path = path_for_level(scenario)
        lua_util.create_script_copy(editor, scenario_path)

        script = self._custom_level_scripts[scenario]

        if not script["edited"]:
            script["script"] += "\nGame.DoFile('actors/items/randomizer_powerup/scripts/randomizer_powerup.lua')\n\n"
            script["edited"] = True

        replacement = {
            "scenario": scenario,
            "funcname": pickup_lua_callback["function"],
            "pickup_class": self.get_script_class(pickup, True),
            "extra_code": extra_code,
            "args": ", ".join([f"_ARG_{i}_" for i in range(pickup_lua_callback["args"])])
        }
        script["script"] += lua_util.replace_lua_template("boss_powerup_template.lua", replacement)

    def patch_corex_pickup_script(self, editor: PatcherEditor, pickup: dict, pickup_lua_callback: dict) -> None:
        bossid = pickup_lua_callback["function"]
        self._corex_replacement[bossid] = self.get_script_class(pickup, True)

    def save_modifications(self, editor: PatcherEditor) -> None:
        editor.replace_asset("actors/items/randomizer_powerup/scripts/randomizer_powerup.lc", self._powerup_script)
        for scenario, script in self._custom_level_scripts.items():
            editor.replace_asset(path_for_level(scenario) + ".lc", script["script"].encode("utf-8"))

        for boss in {"core_x", "core_x_superquetzoa"}:
            corex_script = lua_util.replace_lua_template(f"custom_{boss}.lua", self._corex_replacement).encode("utf-8")
            editor.replace_asset(f"actors/characters/{boss}/scripts/{boss}.lc", corex_script)



import copy

import json_delta

from open_dread_rando.patcher_editor import PatcherEditor


def _modify_actor(editor: PatcherEditor, original_reference: dict[str, str], new_reference: dict[str, str],
                  modifications: list[dict], actor_groups: list[str], make_copy: bool = False):
    is_in_place = original_reference == new_reference
    scenario = editor.get_scenario(new_reference["scenario"])
    new_actor_name = new_reference["actor"]
    new_sublayer = new_reference.get("sublayer", new_reference["layer"])
    new_actor_layer = new_reference["actor_layer"]

    if is_in_place:
        actor = editor.resolve_actor_reference(original_reference)
        make_copy = False
    else:
        actor = copy.deepcopy(editor.resolve_actor_reference(original_reference))
        actor.sName = new_actor_name
        actors_in_sublayer = scenario.actors_for_sublayer(new_sublayer, new_actor_layer)
        actors_in_sublayer[new_actor_name] = actor

    if modifications:
        json_delta.patch(actor, [
            [modification["path"], modification["update_to"]] if "update_to" in modification
            else [modification["path"]]
            for modification in modifications
        ])

    if actor_groups is not None:
        for group in scenario.actor_groups_for_actor_layer(new_actor_layer):

            is_actor_in_group = scenario.is_actor_in_group(group, new_actor_name, new_sublayer, new_actor_layer)

            if group in actor_groups and not is_actor_in_group:
                scenario.add_actor_to_group(group, new_actor_name, new_sublayer, new_actor_layer)
            elif is_actor_in_group:
                scenario.remove_actor_from_group(group, new_actor_name, new_sublayer, new_actor_layer)

    elif (not is_in_place and
          original_reference["scenario"] == new_reference["scenario"] and
          original_reference["actor_layer"] == new_actor_layer):
        editor.copy_actor_groups(original_reference, new_reference, scenario.name)

    if not is_in_place and not make_copy:
        editor.remove_entity(original_reference, None)

def _remove_actor(editor: PatcherEditor, actor: dict):
    editor.remove_entity(actor["actor"], actor.get("map_category", None))

def apply_actor_patches(editor: PatcherEditor, actors_config: dict):
    if "modify" in actors_config:
        for actor in actors_config["modify"]:
            _modify_actor(
                editor,
                actor["actor"],
                actor.get("new_reference", actor["actor"]),
                actor["modifications"],
                actor.get("actor_groups"),
                actor["make_copy"]
            )

    if "remove" in actors_config:
        for actor in actors_config["remove"]:
            _remove_actor(editor, actor)
